// ACS-IBE demo (2-in-1), V1.6, hidraw/libusb-1.0 version, Linux

// _DEMO_MODE_ must be 0 or 1: 0 - IBE-demo, 1 - ACS-demo
#define _DEMO_MODE_         0

#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include "rd0xAB.h"

#define PRG_VERSION         "1.6"

#ifdef _RD0XAB_HIDRAW_H_
#define PRG_COMMENT         " Hidraw version."
#else
#define PRG_COMMENT         " Libusb version."
#endif  // endif for subsytem info


#if _DEMO_MODE_ == 0        // demo mode selector
//#pragma message("*** IBE demo compiling ***")
#define PRG_NAME            "ibe-demo"              // program name
#define DEVM_CN             (2 + 8)                 // ACS(0) or IBE(2) mode + buzz_en(+8)
#define OPTS_CN             (4)                     // rwc_en(+4)+sto_en(+2)+polar(+1)
#define TMRS_CN             ((15 << 4) + 9)         // lnk_tout (0..15 << 4) + on_tout(0..15)
#else
//#pragma message("*** ACS demo compiling ***")
#define PRG_NAME            "acs-demo"              // program name
#define DEVM_CN             (0 + 8)                 // ACS(0) or IBE(2) mode + buzz_en(+8)
#define OPTS_CN             (4)                     // rwc_en(+4)+sto_en(+2)+polar(+1)
#define TMRS_CN             ((15 << 4) + 9)         // lnk_tout (0..15 << 4) + on_tout(0..15)
#endif  // endif for demo mode selector

// convert uidlen, atqa, sak to card type (copied from mif0xAB.c)
const char* acq2type_verb(DWORD acq)
{
    const char*      csType = "UNKN";

    if (acq == 0x07004400) csType = "UL";           // doub, 0044, 00 MIFARE Ultralight CL2
    else if (acq == 0x04000409) csType = "MINI";    // once, 0004, 09 MIFARE Mini

    else if (acq == 0x04000408) csType = "CL1K";    // once, 0004, 08 MIFARE 1K
    else if (acq == 0x04000488) csType = "CL1K";    // once, 0004, 18 MIFARE 1K
    else if (acq == 0x04000218) csType = "CL4K";    // once, 0002, 18 MIFARE 4K

    else if (acq == 0x07000404) csType = "CL1K1";   // once, 0004, 04 Any MIFARE CL1
    else if (acq == 0x07000204) csType = "CL4K1";   // once, 0002, 04 Any MIFARE CL1
    else if (acq == 0x07004404) csType = "CL1K1";   // once, 0044, 04 Any MIFARE CL1
    else if (acq == 0x07004204) csType = "CL4K1";   // once, 0042, 04 Any MIFARE CL1

    else if (acq == 0x07004408) csType = "PL2S1";   // doub, 0044, 08 MIFARE Plus CL2 2K Sec.lev.1
    else if (acq == 0x07004218) csType = "PL4S1";   // doub, 0042, 18 MIFARE Plus CL2 4K Sec.lev.1

    else if (acq == 0x04000410) csType = "PL2S2";   // once, 0004, 10 MIFARE Plus 2K Sec.lev.2
    else if (acq == 0x04000211) csType = "PL4S2";   // once, 0002, 11 MIFARE Plus 4K Sec.lev.2

    else if (acq == 0x07004410) csType = "PL2S2";   // doub, 0044, 10 MIFARE Plus CL2 2K Sec.lev.2
    else if (acq == 0x07004211) csType = "PL4S2";   // doub, 0042, 11 MIFARE Plus CL2 4K Sec.lev.2

    else if (acq == 0x04000420) csType = "PL2S3";   // once, 0004, 20 MIFARE Plus 2K Sec.lev.3
    else if (acq == 0x04000220) csType = "PL4S3";   // once, 0002, 20 MIFARE Plus 4K Sec.lev.3
    else if (acq == 0x07004420) csType = "PL2S3";   // doub, 0044, 20 MIFARE Plus 2K Sec.lev.3
    else if (acq == 0x07004220) csType = "PL4S3";   // doub, 0042, 20 MIFARE Plus 4K Sec.lev.3

    else if (acq == 0x07034420) csType = "DF2";     // doub, 0344, 20 MIFARE DESFire CL2
    else if (acq == 0x07034460) csType = "DF2";     // doub, 0344, 60 MIFARE DESFire card emulation
    else if ((acq&(~0xFF0000FF)) == 0x07000024) csType = "DF1";     // doub, xxxx, 24 MIFARE DESFire CL1

    else if ((acq&(~0x000F0000)) == 0x04000420) csType = "SM1K";    // once, 0x04, 20 SmartMX with MIFARE 1K emulation
    else if ((acq&(~0x000F0000)) == 0x04000428) csType = "SM1K";    // once, 0x04, 28 SmartMX | SAM)AV2 with MIFARE 1K emulation
    else if ((acq&(~0x000F0000)) == 0x04000468) csType = "SM1K";    // once, 0x04, 68 SmartMX with MIFARE 1K emulation

    else if ((acq&(~0x000F0000)) == 0x04000238) csType = "SM4K";    // once, 0x02, 38 SmartMX with MIFARE 4K emulation
    else if ((acq&(~0x000F0000)) == 0x07004820) csType = "SMMX";    // once, 0x48, 20 SmartMX

    return csType;
}

void verbose_err(DWORD err, char* szErr)
{
    if (err == LERR_SUCCESS) strcpy(szErr, "Success!");
    else if (err == LERR_FAIL) strcpy(szErr, "Command execution error!");
    else if (err == LERR_HARDWARE) strcpy(szErr, "Hardware error!");
    else if (err == LERR_TIMEOUT_TX) strcpy(szErr, "Communication error: TX timeout!");
    else if (err == LERR_TIMEOUT_RX) strcpy(szErr, "Communication error: RX timeout!");
    else if (err == LERR_NOTSYNC) strcpy(szErr, "Communication error: synchronization lost!");
    else if (err == LERR_NOREADER) strcpy(szErr, "Can't find reader!");
    else if (err == LERR_NOCARD) strcpy(szErr, "Can't find card!");
    else if (err == LERR_INVALIDCARD) strcpy(szErr, "Invalid card type!");
    else if (err == LERR_INVALIDKEY) strcpy(szErr, "Invalid auth key or card type!");
    else if (err == LERR_IOERR) strcpy(szErr, "I/O error!");
    else if (err == LERR_DATAREAD) strcpy(szErr, "Can't read data from card!");
    else if (err == LERR_DATAWRITE) strcpy(szErr, "Can't write data to card!");
    else if (err == LERR_BUSY) strcpy(szErr, "Device busy error!");
    else strcpy(szErr, "Unknown error!");
}

void usage(void)
{
    printf("Usage for USB: %s\n", PRG_NAME);
    printf("Usage for COM: %s /dev/tty2 9600\n", PRG_NAME);
}

BOOL _kbhit(void)
{
     struct   termios oldt, newt;
     int              ch;
     int              oldf;

     tcgetattr(STDIN_FILENO, &oldt);
     newt = oldt;
     newt.c_lflag &= ~(ICANON | ECHO);
     tcsetattr(STDIN_FILENO, TCSANOW, &newt);
     oldf = fcntl(STDIN_FILENO, F_GETFL, 0);
     fcntl(STDIN_FILENO, F_SETFL, oldf | O_NONBLOCK);

     ch = getchar();

     tcsetattr(STDIN_FILENO, TCSANOW, &oldt);
     fcntl(STDIN_FILENO, F_SETFL, oldf);

     if (ch != EOF)
     {
         ungetc(ch, stdin);
         return TRUE;
     }
     return FALSE;
}

// *** main ***
int  main(int argc, char *const argv[])
{
    CCR             ccr;
    char            *devtty = NULL;
    DWORD           cbr = 0;
    DWORD           err = LERR_SUCCESS;
    BYTE            dm = 0, opt = 0, tmr = 0;

    char            szLocal[32];
    BYTE            buf[14], key;
    DWORD           cnt;
    BOOL            brkf = FALSE;

    memset(&ccr, 0, sizeof(ccr));
    printf("%s COM/USB, V%s.%s\n--\n", PRG_NAME, PRG_VERSION, PRG_COMMENT);
     // Parse command line arguments
    if ((argc != 1) && (argc != 3))
    {
        usage();
        goto main_ferr;
    }
    if (argc == 3)
    {
        devtty = argv[1];
        sscanf(argv[2], "%u", &cbr);
    }
    if ((!devtty) && (argc > 1))
    {
        printf("Invalid parameters!\n\n");
        usage();
        goto main_ferr;
    }

    // if port != 0 then try open COM else try open USB
    if (devtty)
    {
        printf("COM: %s, CBR=%u. ", devtty, cbr);
        if ((err = link_comopen(devtty, cbr, &ccr))) goto main_ferr;
    }
    else
    {
        printf("USB. ");
        if ((err = link_hidopen(0, NULL, &ccr)))
        {
             err = LERR_NOREADER;
             goto main_ferr;
        }
    }
    // read device description: HW, NVMSZ, model, SN
    if ((err = link_packet(&ccr, CMG_VER, 0, NULL, 0, buf, 6)) != LERR_SUCCESS)
    {
         if ((devtty) && (err == LERR_HARDWARE)) err = LERR_NOREADER;
         goto main_ferr;
    }
    printf("Reader ");
    if ((buf[2] & 0xF0) == 0x10) printf("RD-01A");
    else if ((buf[2] & 0xF0) > 0x10) printf("RD-%02uAB", buf[2] / 16);
    else printf("???");
    printf(", SN:%08X, HW:%u.%u, %uKb.",
        (((DWORD)buf[2]) << 24) | (((DWORD)buf[3]) << 16) | (((DWORD)buf[4]) << 8) | ((DWORD)buf[5]),
        buf[0] >> 4, buf[0] & 15, ((DWORD)buf[1] + 1) / 2);
    if (buf[0] < 0x20)
    {
        printf(" ERROR: HW version must be >= 2.0!\n");
        goto main_ferr;
    }
    printf("\n--\n");

    // temporary set DIR mode, buzzer on
    if ((err = link_packet(&ccr, CMS_MOD, 3 + 8, NULL, 0, NULL, 0)) != LERR_SUCCESS) goto main_ferr;

    // *** ATS/IBE mode settings ***
    if ((DEVM_CN & 3) == 0) strcpy(szLocal, "ACS");
    else if ((DEVM_CN & 3) == 2) strcpy(szLocal, "IBE");
    else sprintf(szLocal, "%u - invalid!", DEVM_CN & 3);
    printf("Mode .............. %s.\n", szLocal);
    if (((DEVM_CN & 3) != 0) && ((DEVM_CN & 3) != 2)) goto main_ferr;
    printf("Buzzer ............ %s.\n", (DEVM_CN & 8) ? "enable" : "disable");
    printf("RW-card ........... %s.\n", (OPTS_CN & 4) ? "enable" : "disable");
    printf("Polarity .......... %s.\n", (OPTS_CN & 1) ? "inverted" : "direct");
    printf("Ack timeout ....... %.1fs.\n", (float)((TMRS_CN & 0xF0) | 15) * 25 / 1000);             // time = tmr_nibble * 16 * 25mS
    printf("Unlock time ....... %.1fs.\n\n", (float)(((TMRS_CN & 0x0F) << 4) | 15) * 50 / 1000);    // time = tmr_nibble * 16 * 50mS

    printf("--\nSetup ACS/IBE mode... ");
    // *** set if difference only ***
    // options
    if ((err = link_packet(&ccr, CMG_OPT, 0x80, NULL, 0, &opt, 1)) != LERR_SUCCESS) goto main_ferr;
    if ((opt != OPTS_CN) && ((err = link_packet(&ccr, CMS_OPT, OPTS_CN, NULL, 0, NULL, 0)) != LERR_SUCCESS)) goto main_ferr;
    // timers
    if ((err = link_packet(&ccr, CMG_TMR, 0, NULL, 0, &tmr, 1)) != LERR_SUCCESS) goto main_ferr;
    if ((tmr != TMRS_CN) && ((err = link_packet(&ccr, CMS_TMR, TMRS_CN, NULL, 0, NULL, 0)) != LERR_SUCCESS)) goto main_ferr;
    // device mode
    if ((err = link_packet(&ccr, CMG_MOD, 0x80, NULL, 0, &dm, 1)) != LERR_SUCCESS) goto main_ferr;
    if ((err = link_packet(&ccr, CMS_MOD, DEVM_CN | ((dm != DEVM_CN) ? 0x80 : 0), NULL, 0, NULL, 0)) != LERR_SUCCESS) goto main_ferr;
    // final beep :-)
    if ((err = link_packet(&ccr, CMF_BUZ, 0x02, NULL, 0, NULL, 0)) != LERR_SUCCESS) goto main_ferr;
    printf("Ok.\n");

    // *** main loop ***
    printf("--\nPress 1 for lock, 2 for unlock, 3 for default action (int. db in ACS-mode).\nPress Esc for exit.\n\nWait for card...\n");
    while (!brkf)
    {
        time_t          systm;
        struct   tm     *ptm;
        DWORD           stoptick;

        if (_kbhit())
        {
            if ((key = (BYTE)getchar()) == 0) key = (BYTE)getchar();
            if (key == 0x1B)
            {
                printf("^Esc pressed!\n");
                brkf = TRUE;
                break;
            }
        }
        // wait for unlock request from reader: length = 14b{1b(uidlen) + 10b(uid) + 1b(sak) + 2b(atqa)}
        if ((err = link_rqwait(&ccr, buf, 14)) == LERR_HARDWARE) goto main_ferr;
        if ((err != LERR_SUCCESS) || ((buf[0] & 0x0F) > 10)) continue;  // uidlen <= 10, b.7 - RW-flag

        // build type string
        strcpy(szLocal, acq2type_verb( ((DWORD)(buf[0] & 0x0F) << 24) |  // uidlen
            (((((DWORD)buf[12]) << 8) | ((DWORD)buf[13])) << 8) | ((DWORD)buf[11])));   // ATQA + SAK
        if (buf[0] & MFDM_RWC) strcat(szLocal, "-RW");
        // print request message
        time(&systm); ptm = localtime(&systm);
        printf("[%2u/%02u/%02u %2u:%02u:%02u] %8s",
            ptm->tm_mday, ptm->tm_mon + 1, ptm->tm_year % 100,
            ptm->tm_hour, ptm->tm_min, ptm->tm_sec,
            szLocal);   // print time, card type
        // build UID string
        for (szLocal[0] = 0, cnt = 0; cnt < (DWORD)(buf[0] & 0x0F); cnt++)
            sprintf(&szLocal[strlen(szLocal)], "%02X", buf[cnt + 1]);
        printf(", UID:%14s. Unlock? (1|2|3): ", szLocal);   // print UID
        fflush(stdout);
        // loop for getting acknowledge from user
        key = 0;
        stoptick = GetTickCount() + (TMRS_CN & 0xF0) * 25;  // count timeout time
        while (GetTickCount() < stoptick)
        {
            if (_kbhit())
            {
                if ((key = (BYTE)getchar()) == 0) key = (BYTE)getchar();
                if ((key == '1') || (key == '2') || (key == '3') || (key == 0x1B)) break;
            }
        }
        if (key == 0x1B) brkf = TRUE;
        if ((key != '1') && (key != '2') && (key != '3')) printf(" none.\n");
        else
        {
            if (key == '1') printf(" lock");
            else if (key == '2') printf(" unlk");
            else printf(" dflt");
            printf("...");
            key = (key & 3) - 1;
            // *** sending ack ***
            // 0 - no, 1 - yes, 2 - by reader db
            if ((err = link_rqack(&ccr, key)) == LERR_HARDWARE) goto main_ferr;
            if (err == LERR_SUCCESS) printf(" Ok.\n");
            else printf(" Fail.\n");
        }
    }
main_ferr:
    link_close(&ccr);
    if ((err) && (!brkf))
    {
        verbose_err(err, szLocal);
        printf("%s\n", szLocal);
    }
    return 0;
}
