/*  crapto1.h

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
    MA  02110-1301, US$

    Copyright (C) 2008-2009 bla <blapost@gmail.com>
*/
#ifndef CRAPTO1_INCLUDED
#define CRAPTO1_INCLUDED
#include "rd0xAB.h"
typedef long long unsigned int  DWORD64;
typedef long long unsigned int* PDWORD64;

#ifdef __cplusplus
extern "C" {
#endif

  struct Crypto1State {DWORD odd, even;};
  struct Crypto1State *crypto1_create(DWORD64);
  void crypto1_destroy(struct Crypto1State *);
  void crypto1_get_lfsr(struct Crypto1State *, DWORD64 *);
  BYTE crypto1_bit(struct Crypto1State *, BYTE, int);
  BYTE crypto1_byte(struct Crypto1State *, BYTE, int);
  DWORD crypto1_word(struct Crypto1State *, DWORD, int);
  DWORD prng_successor(DWORD x, DWORD n);

  struct Crypto1State *lfsr_recovery32(DWORD ks2, DWORD in);
  struct Crypto1State *lfsr_recovery64(DWORD ks2, DWORD ks3);

  void lfsr_rollback(struct Crypto1State *s, DWORD in, int fb);
  DWORD lfsr_rollback_word(struct Crypto1State *s, DWORD in, int fb);
  int nonce_distance(DWORD from, DWORD to);
#define FOREACH_VALID_NONCE(N, FILTER, FSIZE)\
  DWORD __n = 0,__M = 0, N = 0;\
  int __i;\
  for(; __n < 1 << 16; N = prng_successor(__M = ++__n, 16))\
    for(__i = FSIZE - 1; __i >= 0; __i--)\
      if(BIT(FILTER, __i) ^ parity(__M & 0xFF01))\
        break;\
      else if(__i)\
        __M = prng_successor(__M, (__i == 7) ? 48 : 8);\
      else

#define LF_POLY_ODD (0x29CE5C)
#define LF_POLY_EVEN (0x870804)
#define BIT(x, n) ((x) >> (n) & 1)
#define BEBIT(x, n) BIT(x, (n) ^ 24)
static inline int parity(DWORD x)
{
    x ^= x >> 16;
    x ^= x >> 8;
    x ^= x >> 4;
    return BIT(0x6996, x & 0xf);
}

static inline int filter(DWORD const x)
{
    DWORD f;

    f  = 0xf22c0 >> (x       & 0xf) & 16;
    f |= 0x6c9c0 >> (x >>  4 & 0xf) &  8;
    f |= 0x3c8b0 >> (x >>  8 & 0xf) &  4;
    f |= 0x1e458 >> (x >> 12 & 0xf) &  2;
    f |= 0x0d938 >> (x >> 16 & 0xf) &  1;
    return BIT(0xEC57E80A, f);
}
#ifdef __cplusplus
}
#endif
#endif
