// RQV (ReQuest Viewer) demo, V1.6, libusb-1.0 version, Linux
// print request (CD - Card Descriptor, 14 bytes): uidlen(1) + uid(10) + sak(1) + atqa(2)

#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include "rd0xAB.h"

#define PRG_NAME            "rqv"       // program name
#define PRG_VERSION         "1.6"

#ifdef _RD0XAB_HIDRAW_H_
#define PRG_COMMENT         " Hidraw version."
#else
#define PRG_COMMENT         " Libusb version."
#endif  // endif for subsytem info


// convert uidlen, atqa, sak to card type (copied from mif0xAB.c)
const char* acq2type_verb(DWORD acq)
{
    const char*      csType = "UNKN";

    if (acq == 0x07004400) csType = "UL";           // doub, 0044, 00 MIFARE Ultralight CL2
    else if (acq == 0x04000409) csType = "MINI";    // once, 0004, 09 MIFARE Mini

    else if (acq == 0x04000408) csType = "CL1K";    // once, 0004, 08 MIFARE 1K
    else if (acq == 0x04000488) csType = "CL1K";    // once, 0004, 18 MIFARE 1K
    else if (acq == 0x04000218) csType = "CL4K";    // once, 0002, 18 MIFARE 4K

    else if (acq == 0x07000404) csType = "CL1K1";   // once, 0004, 04 Any MIFARE CL1
    else if (acq == 0x07000204) csType = "CL4K1";   // once, 0002, 04 Any MIFARE CL1
    else if (acq == 0x07004404) csType = "CL1K1";   // once, 0044, 04 Any MIFARE CL1
    else if (acq == 0x07004204) csType = "CL4K1";   // once, 0042, 04 Any MIFARE CL1

    else if (acq == 0x07004408) csType = "PL2S1";   // doub, 0044, 08 MIFARE Plus CL2 2K Sec.lev.1
    else if (acq == 0x07004218) csType = "PL4S1";   // doub, 0042, 18 MIFARE Plus CL2 4K Sec.lev.1

    else if (acq == 0x04000410) csType = "PL2S2";   // once, 0004, 10 MIFARE Plus 2K Sec.lev.2
    else if (acq == 0x04000211) csType = "PL4S2";   // once, 0002, 11 MIFARE Plus 4K Sec.lev.2

    else if (acq == 0x07004410) csType = "PL2S2";   // doub, 0044, 10 MIFARE Plus CL2 2K Sec.lev.2
    else if (acq == 0x07004211) csType = "PL4S2";   // doub, 0042, 11 MIFARE Plus CL2 4K Sec.lev.2

    else if (acq == 0x04000420) csType = "PL2S3";   // once, 0004, 20 MIFARE Plus 2K Sec.lev.3
    else if (acq == 0x04000220) csType = "PL4S3";   // once, 0002, 20 MIFARE Plus 4K Sec.lev.3
    else if (acq == 0x07004420) csType = "PL2S3";   // doub, 0044, 20 MIFARE Plus 2K Sec.lev.3
    else if (acq == 0x07004220) csType = "PL4S3";   // doub, 0042, 20 MIFARE Plus 4K Sec.lev.3

    else if (acq == 0x07034420) csType = "DF2";     // doub, 0344, 20 MIFARE DESFire CL2
    else if (acq == 0x07034460) csType = "DF2";     // doub, 0344, 60 MIFARE DESFire card emulation
    else if ((acq&(~0xFF0000FF)) == 0x07000024) csType = "DF1";     // doub, xxxx, 24 MIFARE DESFire CL1

    else if ((acq&(~0x000F0000)) == 0x04000420) csType = "SM1K";    // once, 0x04, 20 SmartMX with MIFARE 1K emulation
    else if ((acq&(~0x000F0000)) == 0x04000428) csType = "SM1K";    // once, 0x04, 28 SmartMX | SAM)AV2 with MIFARE 1K emulation
    else if ((acq&(~0x000F0000)) == 0x04000468) csType = "SM1K";    // once, 0x04, 68 SmartMX with MIFARE 1K emulation

    else if ((acq&(~0x000F0000)) == 0x04000238) csType = "SM4K";    // once, 0x02, 38 SmartMX with MIFARE 4K emulation
    else if ((acq&(~0x000F0000)) == 0x07004820) csType = "SMMX";    // once, 0x48, 20 SmartMX

    return csType;
}

void verbose_err(DWORD err, char* szErr)
{
    if (err == LERR_SUCCESS) strcpy(szErr, "Success!");
    else if (err == LERR_FAIL) strcpy(szErr, "Command execution error!");
    else if (err == LERR_HARDWARE) strcpy(szErr, "Hardware error!");
    else if (err == LERR_TIMEOUT_TX) strcpy(szErr, "Communication error: TX timeout!");
    else if (err == LERR_TIMEOUT_RX) strcpy(szErr, "Communication error: RX timeout!");
    else if (err == LERR_NOTSYNC) strcpy(szErr, "Communication error: synchronization lost!");
    else if (err == LERR_NOREADER) strcpy(szErr, "Can't find reader!");
    else if (err == LERR_NOCARD) strcpy(szErr, "Can't find card!");
    else if (err == LERR_INVALIDCARD) strcpy(szErr, "Invalid card type!");
    else if (err == LERR_INVALIDKEY) strcpy(szErr, "Invalid auth key or card type!");
    else if (err == LERR_IOERR) strcpy(szErr, "I/O error!");
    else if (err == LERR_DATAREAD) strcpy(szErr, "Can't read data from card!");
    else if (err == LERR_DATAWRITE) strcpy(szErr, "Can't write data to card!");
    else if (err == LERR_BUSY) strcpy(szErr, "Device busy error!");
    else strcpy(szErr, "Unknown error!");
}

void usage(void)
{
    printf("Usage for USB: %s\n", PRG_NAME);
    printf("Usage for COM: %s /dev/tty2 9600\n", PRG_NAME);
}

BOOL _kbhit(void)
{
     struct   termios oldt, newt;
     int              ch;
     int              oldf;

     tcgetattr(STDIN_FILENO, &oldt);
     newt = oldt;
     newt.c_lflag &= ~(ICANON | ECHO);
     tcsetattr(STDIN_FILENO, TCSANOW, &newt);
     oldf = fcntl(STDIN_FILENO, F_GETFL, 0);
     fcntl(STDIN_FILENO, F_SETFL, oldf | O_NONBLOCK);

     ch = getchar();

     tcsetattr(STDIN_FILENO, TCSANOW, &oldt);
     fcntl(STDIN_FILENO, F_SETFL, oldf);

     if (ch != EOF)
     {
         ungetc(ch, stdin);
         return TRUE;
     }
     return FALSE;
}

// *** main ***
int  main(int argc, char *const argv[])
{
    CCR             ccr;
    char            *devtty = NULL;
    DWORD           cbr = 0;
    DWORD           err = LERR_SUCCESS;

    char            szLocal[32];
    BYTE            buf[14], key;

    BYTE            sc_dm = 0, sc_opts = 0, sc_tmrs = 0;
    BYTE            sc_aksd[4] = { 0, 0, 0, 0 };
    DWORD           cnt, tmp;
    BOOL            brkf = FALSE;

    memset(&ccr, 0, sizeof(ccr));
    printf("RQV - ReQuest Viewer demo, COM/USB, V%s.%s\n--\n", PRG_VERSION, PRG_COMMENT);
    // Parse command line arguments
    if ((argc != 1) && (argc != 3))
    {
        usage();
        goto main_ferr;
    }
    if (argc == 3)
    {
        devtty = argv[1];
        sscanf(argv[2], "%u", &cbr);
    }
    if ((!devtty) && (argc > 1))
    {
        printf("Invalid parameters!\n\n");
        usage();
        goto main_ferr;
    }

    // if port != 0 then try open COM else try open USB
    if (devtty)
    {
        printf("COM: %s, CBR=%u. ", devtty, cbr);
        if ((err = link_comopen(devtty, cbr, &ccr))) goto main_ferr;
    }
    else
    {
        printf("USB. ");
        if ((err = link_hidopen(0, NULL, &ccr)))
        {
             err = LERR_NOREADER;
             goto main_ferr;
        }
    }
    // read device description: HW, NVMSZ, model, SN
    if ((err = link_packet(&ccr, CMG_VER, 0, NULL, 0, buf, 6)) != LERR_SUCCESS)
    {
         if ((devtty) && (err == LERR_HARDWARE)) err = LERR_NOREADER;
         goto main_ferr;
    }
    printf("Reader ");
    if ((buf[2] & 0xF0) == 0x10) printf("RD-01A");
    else if ((buf[2] & 0xF0) > 0x10) printf("RD-%02uAB", buf[2] / 16);
    else printf("???");
    printf(", SN:%08X, HW:%u.%u, %uKb.",
        (((DWORD)buf[2]) << 24) | (((DWORD)buf[3]) << 16) | (((DWORD)buf[4]) << 8) | ((DWORD)buf[5]),
        buf[0] >> 4, buf[0] & 15, ((DWORD)buf[1] + 1) / 2);
    if (buf[0] < 0x20)
    {
        printf(" ERROR: HW version must be >= 2.0!\n");
        goto main_ferr;
    }
    printf("\n--\n");

    // temporary set DIR mode, buzzer on
    if ((err = link_packet(&ccr, CMS_MOD, 3 + 8, NULL, 0, NULL, 0)) != LERR_SUCCESS) goto main_ferr;

    // *** read ROM settings ***
    // buzzer & device mode
    if ((err = link_packet(&ccr, CMG_MOD, 0x80, NULL, 0, &sc_dm, 1)) != LERR_SUCCESS) goto main_ferr;
    // options
    if ((err = link_packet(&ccr, CMG_OPT, 0x80, NULL, 0, &sc_opts, 1)) != LERR_SUCCESS) goto main_ferr;
    // timers
    if ((err = link_packet(&ccr, CMG_TMR, 0, NULL, 0, &sc_tmrs, 1)) != LERR_SUCCESS) goto main_ferr;
    // SRD parameters
    if ((err = link_packet(&ccr, CMG_SRD, 0, NULL, 0, sc_aksd, 4)) != LERR_SUCCESS) goto main_ferr;
    // restore device mode (ROM -> RAM)
    if ((err = link_packet(&ccr, CMS_MOD, sc_dm, NULL, 0, NULL, 0)) != LERR_SUCCESS) goto main_ferr;


    // *** print & parse readen parameters ***
    brkf = FALSE;   // TRUE if invalid settings
    // device mode
    printf("Mode ................... ");
    if ((sc_dm & 3) == 0) printf("ACS");
    else if ((sc_dm & 3) == 1) printf("SRD");
    else if ((sc_dm & 3) == 2) printf("IBE");
    else
    {
        printf("DIR (invalid for 'request only')");
        brkf = TRUE;
    }
    printf("\n");
    // buzzer
    printf("Buzzer ................. %s.\n", (sc_dm & 8) ? "enable" : "disable");
    // RW-flag
    printf("RW-card ................ %s.\n", (sc_opts & 4) ? "enable" : "disable");
    // polarity
    printf("Polarity ............... %s.\n", (sc_opts & 1) ? "inverted" : "direct");
    // response timeout
    printf("Response timeout ....... ");
    tmp = (DWORD)(sc_tmrs & 0xF0) >> 4;     // select link_timer nibble
    if (tmp == 0)
    {
        printf("without request (invalid for 'request only'.\n");
    }
    else if (tmp == 1) printf("request only.\n");
    else
    {
        // time = (ack_timer * 16 + 15) * 25mS
        printf("%.1fs (incorrect for 'request only').\n", (float)(tmp * 16 + 15) * 25 / 1000);
    }
    // unlock time
    printf("Unlock time ............ ");
    tmp = (DWORD)(sc_tmrs & 0x0F);          // select unlock_timer nibble
    if (tmp == 0)
    {
        printf("trigger mode.\n");
    }
    else
    {
        // time = (unlock_timer * 16 + 15) * 50mS
        printf("%.1fs.\n", (float)(tmp * 16 + 15) * 50 / 1000);
    }

    // authorization command
    printf("\nAuth command ........... %02X", sc_aksd[0]);
    if (((sc_dm & 3) == 1) && (sc_aksd[0] == 0))
    {
        printf(" (invalid for 'request only').\n");
        brkf = TRUE;
    }
    else printf(".\n");
    // authorization key number (AK)
    printf("Auth key number ........ %u", sc_aksd[1]);
    if (((sc_dm & 3) == 1) && ((sc_aksd[1] == 0) || (sc_aksd[1] > 84)))
    {
        printf(" (invalid key number for SRD mode).\n");
        brkf = TRUE;
    }
    else printf(".\n");
    // start block/page
    printf("Start block/page ....... %u.\n", sc_aksd[2]);
    // auth/data read mode
    printf("Data/Auth mode ......... %s", (sc_aksd[3] & 2) ? "auth only" : "data read");
    if (((sc_dm & 3) == 1) && ((sc_aksd[3] & 2) == 0))
    {
        printf(" (invalid for 'request only demo').\n");
        brkf = TRUE;
    }
    else printf(".\n");
    // page mode
    printf("Block/page mode ........ %s.\n", (sc_aksd[3] & 1) ? "page mode (4 bytes per unit)" : "block (16 bytes per unit)");

    // block's/page's counter : datalen = (sc_aksd[3] & 2) ? 0 : ((sc_aksd[3] & 1) ? 4 : 16) * ((sc_aksd[3] >> 4) + 1);
    if (sc_aksd[3] & 2) tmp = 0;
    else if (sc_aksd[3] & 1) tmp = 4 * ((sc_aksd[3] >> 4) + 1);
    else tmp = 16 * ((sc_aksd[3] >> 4) + 1);
    printf("Blocks/pages for read .. %u (%u bytes).\n", (sc_aksd[3] >> 4) + 1, tmp);
    printf("--\n");
    // exit if invalid parameters
    if (brkf)
    {
        printf("Can't continue with current parameters! Change options then re-run program.\n");
        goto main_ferr;
    }

    // *** main loop for view request data ***
    printf("Press Esc for exit.\n\nWait for card...\n");
    while (!brkf)
    {
        time_t          systm;
        struct   tm     *ptm;
        DWORD           stoptick;

        if (_kbhit())
        {
            if ((key = (BYTE)getchar()) == 0) key = (BYTE)getchar();
            if (key == 0x1B)
            {
                printf("^Esc pressed!\n");
                brkf = TRUE;
                break;
            }
        }
        // wait for unlock request from reader: length = 14b{1b(uidlen) + 10b(uid) + 1b(sak) + 2b(atqa)}
        if ((err = link_rqwait(&ccr, buf, SZ_SCCD)) == LERR_HARDWARE) goto main_ferr;
        if ((err != LERR_SUCCESS) || ((buf[0] & 0x0F) > SZ_UID)) continue;  // uidlen <= 10, b.7 - RW-flag

        // build type string
        strcpy(szLocal, acq2type_verb( ((DWORD)(buf[0] & 0x0F) << 24) |                 // uidlen
            (((DWORD)buf[12]) << 16) | (((DWORD)buf[13]) << 8) | ((DWORD)buf[11])));    // ATQA + SAK
        if (buf[0] & MFDM_RWC) strcat(szLocal, "-RW");
        // print request message
        time(&systm); ptm = localtime(&systm);
        printf("[%2u/%02u/%02u %2u:%02u:%02u] %8s",
            ptm->tm_mday, ptm->tm_mon + 1, ptm->tm_year % 100,
            ptm->tm_hour, ptm->tm_min, ptm->tm_sec,
            szLocal);   // print time, card type
        // build UID string
        for (szLocal[0] = 0, cnt = 0; cnt < (DWORD)(buf[0] & 0x0F); cnt++)
            sprintf(&szLocal[strlen(szLocal)], "%02X", buf[cnt + 1]);
        printf(", UID:%14s.\n", szLocal);   // print UID
        // loop for getting acknowledge from user
        key = 0;
        stoptick = GetTickCount() + ((sc_tmrs & 0xF0) + 15) * 25;   // count timeout: link_time = (n * 16 + 15) * 25 mS
        while (GetTickCount() < stoptick)
        {
            if (_kbhit())
            {
                if ((key = (BYTE)getchar()) == 0) key = (BYTE)getchar();
                if (key == 0x1B) break;
            }
        }
        if (key == 0x1B) brkf = TRUE;
    }
main_ferr:
    link_close(&ccr);
    if ((err) && (!brkf))
    {
        verbose_err(err, szLocal);
        printf("%s\n", szLocal);
    }
    return 0;
}
