// scprobe: Simple contactless card analyzer, V1.6 hidraw/libusb-1.0, Linux
#include "mf0xAB.h"

#define PRG_VERSION         "1.6"
#ifdef _RD0XAB_HIDRAW_H_
#define PRG_COMMENT         " Hidraw version."
#else
#define PRG_COMMENT         " Libusb version."
#endif  // endif for subsytem info

void verbose_err(DWORD err, char* szErr)
{
    if (err == LERR_SUCCESS) strcpy(szErr, "Success!");
    else if (err == LERR_FAIL) strcpy(szErr, "Command execution error!");
    else if (err == LERR_HARDWARE) strcpy(szErr, "Hardware error!");
    else if (err == LERR_TIMEOUT_TX) strcpy(szErr, "Communication error: TX timeout!");
    else if (err == LERR_TIMEOUT_RX) strcpy(szErr, "Communication error: RX timeout!");
    else if (err == LERR_NOTSYNC) strcpy(szErr, "Communication error: synchronization lost!");
    else if (err == LERR_NOREADER) strcpy(szErr, "Can't find reader!");
    else if (err == LERR_NOCARD) strcpy(szErr, "Can't find card!");
    else if (err == LERR_INVALIDCARD) strcpy(szErr, "Invalid card type!");
    else if (err == LERR_INVALIDKEY) strcpy(szErr, "Invalid auth key or card type!");
    else if (err == LERR_IOERR) strcpy(szErr, "I/O error!");
    else if (err == LERR_DATAREAD) strcpy(szErr, "Can't read data from card!");
    else if (err == LERR_DATAWRITE) strcpy(szErr, "Can't write data to card!");
    else if (err == LERR_BUSY) strcpy(szErr, "Device busy error!");
    else strcpy(szErr, "Unknown error!");
}

void scp_verbscd(BYTE scd, char* sztp)
{
    if (scd == SCD_UL) strcpy(sztp, "UL");
    else if (scd == SCD_CL1K) strcpy(sztp, "CL1K");
    else if (scd == SCD_CL4K) strcpy(sztp, "CL4K");
    else if (scd == SCD_CL1K1) strcpy(sztp, "CL1K1");
    else if (scd == SCD_CL4K1) strcpy(sztp, "CL4K1");
    else if (scd == SCD_PL2S1) strcpy(sztp, "PL2S1");
    else if (scd == SCD_PL4S1) strcpy(sztp, "PL4S1");
    else if (scd == SCD_PL2S2) strcpy(sztp, "PL2S2");
    else if (scd == SCD_PL4S2) strcpy(sztp, "PL4S2");
    else if (scd == SCD_PL2S3) strcpy(sztp, "PL2S3");
    else if (scd == SCD_PL4S3) strcpy(sztp, "PL4S3");
    else if (scd == SCD_MINI) strcpy(sztp, "MINI");
    else if (scd == SCD_DF1) strcpy(sztp, "DF1");
    else if (scd == SCD_DF2) strcpy(sztp, "DF2");
    else if (scd == SCD_SMMX) strcpy(sztp, "SMMX");
    else if (scd == SCD_SM1K) strcpy(sztp, "SM1K");
    else if (scd == SCD_SM4K) strcpy(sztp, "SM4K");
    else strcpy(sztp, "???");
}

// *** main ***
int  main(int argc, char *const argv[])
{
    CCR         ccr;
    DWORD       err = LERR_SUCCESS;
    char        szLocal[256];
    DWORD       scp_mode = 1, scp_parm1 = 0, scp_parm2 = 0;;

    BYTE        r_model = 0, r_hw = 0, phdl = 0;
    DWORD       r_sn = 0;

    BYTE        sc_type = SCD_UNKNOWN;
    BYTE        sc_uidlen = 0, sc_uid[SZ_UID];
    WORD        sc_atqa = 0;
    BYTE        sc_sak = 0;
    BYTE        sc_atslen, sc_ats[SZ_ATS];

    BYTE        txbuf[128], rxbuf[128], bitlen, bytelen;
    DWORD       cnt, ecnt, idx, rpt;

    printf("SC Probe, contactless card analyzer, V%s.%s\n--\n", PRG_VERSION, PRG_COMMENT);
    memset(&ccr, 0, sizeof(ccr));

    if (argc > 1)
    {
        if (!strcasecmp(argv[1], "-WUPA")) scp_mode = 0;
        else if ((!strcasecmp(argv[1], "-COMM")) && (argc == 3) && (isxdigit(argv[2][0])))
        {
            scp_mode = 1;
            sscanf(argv[2], "%X", &scp_parm1);
            if (scp_parm1 > 255)
            {
                printf("Invalid -COMM parameter!\n");
                goto main_exit;
            }
        }
        else if ((!strcasecmp(argv[1], "-C2PL")) && (argc == 4) && (isxdigit(argv[2][0])) && (isxdigit(argv[3][0])))
        {
            scp_mode = 2;
            sscanf(argv[2], "%X", &scp_parm1);
            if (scp_parm1 > 255)
            {
                printf("Invalid -C2PL parameter 1!\n");
                goto main_exit;
            }
            sscanf(argv[3], "%X", &scp_parm2);
            if (scp_parm2 > 255)
            {
                printf("Invalid -C2PL parameter 2!\n");
                goto main_exit;
            }
        }
        else
        {
            printf("Usage: scprobe -WUPA\n       scprobe -COMM <hex_cmd>\n       scprobe -C2PL <hex_cmd> <hex_parm>\n");
            goto main_exit;
        }
    }

    // open USB device; return LERR_SUCCESS, LERR_HARDWARE
    if ((err = link_hidopen(0, NULL, &ccr)) != LERR_SUCCESS)
    {
        err = LERR_NOREADER;
        goto main_exit;
    }
    // DIR-mode initialization
    if ((err = mfInit(&ccr, &r_model, &r_hw, &r_sn)) != LERR_SUCCESS) goto main_exit;
    printf("RD-0%uAB, SN:%08X, HW:%u.%u\n--\n", r_model, r_sn, r_hw >> 4, r_hw & 15);

    // card type detection
    printf("Search for the type of card... ");
    if ((err = mfType(&ccr, 7, &sc_type, &sc_uidlen, sc_uid, &sc_atqa, &sc_sak, &sc_atslen, sc_ats)) != LERR_SUCCESS) goto main_exit;
    printf("Ok.\n\n");

    scp_verbscd(sc_type, szLocal);
    if (sc_uidlen & MFDM_RWC) strcat(szLocal, "-RW");

    strcat(szLocal, ", UID=");
    for (cnt = 0; (cnt < ((DWORD)sc_uidlen&15)) && (cnt < 10); cnt++)
        sprintf(&szLocal[strlen(szLocal)], "%02X", sc_uid[cnt]);
    sprintf(&szLocal[strlen(szLocal)], ", ATQA=%04X, SAK=%02X.", sc_atqa, sc_sak);
    printf("%s\n", szLocal);

    strcpy(szLocal, "ATS=");
    if (sc_atslen == 0) strcat(szLocal, "none");
    else if (sc_atslen == 1) sprintf(&szLocal[strlen(szLocal)], "none (ack=%02X)", sc_ats[0]);
    else
    {
        for (cnt = 0; ((cnt < sc_atslen) && (cnt < SZ_ATS)); cnt++)
            sprintf(&szLocal[strlen(szLocal)], "%02X", sc_ats[cnt]);
    }
    strcat(szLocal, ".");
    printf("%s\n--\n", szLocal);

    // *** read phd level and extended error flags after select ***
    if ((err = mfSelect(&ccr, 7, PICC_REQALL, NULL, NULL, NULL, NULL)) != LERR_SUCCESS) goto main_exit;
    if ((err = link_packet(&ccr, CMG_ISO, 0, NULL, 0, rxbuf, 2)) != LERR_SUCCESS) goto main_exit;
    phdl = rxbuf[1] & 0x1F;
    printf("PHD level: %02X.\n--\n", phdl);

    if (scp_mode == 0)
    {
        // *** search for WUPA command ***
        printf("WUPA probe loop: from 0 to 127 (7 bits).\n");
        for (ecnt = 0, cnt = 0; cnt != 128; cnt++)
        {
            printf("\r Probe %02X... ", cnt);
            fflush(stdout);
            for (rpt = 3; rpt != 0; rpt--)
            {
                // break carrier
                if ((err = mfBrkCarrier(&ccr, 30)) != LERR_SUCCESS) goto main_exit;
                // set phd level
                if ((err = link_packet(&ccr, CMS_ISO, phdl, NULL, 0, NULL, 0)) != LERR_SUCCESS) goto main_exit;
                // set WUPA command
                txbuf[0] = (BYTE)cnt;
                if (((err = mfBitIO(&ccr, txbuf, 7, rxbuf, &bitlen, NULL)) != LERR_SUCCESS) && (err != LERR_FAIL)) goto main_exit;
                if ((err == LERR_SUCCESS) && (bitlen))
                {
                    ecnt++;
                    bytelen = (bitlen % 8) ? bitlen / 8 + 1 : bitlen / 8;
                    strcpy(szLocal, "ATQA=");
                    for (idx = 0; idx < bytelen; idx++)
                    {
                        if (idx) sprintf(&szLocal[strlen(szLocal)], ".");
                        sprintf(&szLocal[strlen(szLocal)], "%02X", rxbuf[idx]);
                    }
                    sprintf(&szLocal[strlen(szLocal)], " - %u bits.", bitlen);
                    printf("%s\n", szLocal);
                    if ((err = mfBeep(&ccr, 0x02)) != LERR_SUCCESS) goto main_exit;
                    break;
                }
            }
        }
        printf("\rWUPA probe loop completed.  %u acks found.\n", ecnt);
    }
    else if (scp_mode == 1)
    {
        // *** search for command ***
        printf("Command probe loop: from 0 to 255 (8 bits). Parameter: %02X.\n", scp_parm1);
        for (ecnt = 0, cnt = 0; cnt != 256; cnt++)
        {
            printf("\r Probe %02X... ", cnt);
            fflush(stdout);
            for (rpt = 3; rpt != 0; rpt--)
            {
                // break carrier
                if ((err = mfBrkCarrier(&ccr, 30)) != LERR_SUCCESS) goto main_exit;
                // set phd level
                if ((err = link_packet(&ccr, CMS_ISO, phdl | 0x20, NULL, 0, NULL, 0)) != LERR_SUCCESS) goto main_exit;
                // select card
                if ((err = mfSelect(&ccr, 7, PICC_REQALL, NULL, NULL, NULL, NULL)) != LERR_SUCCESS) goto main_exit;
                // set command
                txbuf[0] = (BYTE)cnt; txbuf[1] = (BYTE)scp_parm1;
                rxbuf[0] = 0;
                if (((err = mfLong(&ccr, txbuf, 2, rxbuf, &bitlen)) != LERR_SUCCESS) && (err != LERR_FAIL)) goto main_exit;
                if ((err == LERR_SUCCESS) && ((bitlen > 4) || (rxbuf[0] == 0x0A) ))
                {
                    ecnt++;
                    bytelen = (bitlen % 8) ? bitlen / 8 + 1 : bitlen / 8;
                    strcpy(szLocal, "ACK=");
                    for (idx = 0; idx < bytelen; idx++)
                    {
                        if (idx) sprintf(&szLocal[strlen(szLocal)], ".");
                        sprintf(&szLocal[strlen(szLocal)], "%02X", rxbuf[idx]);
                    }
                    sprintf(&szLocal[strlen(szLocal)], "|%u bits", bitlen);
                    printf("%s\n", szLocal);
                    if ((err = mfBeep(&ccr, 0x02)) != LERR_SUCCESS) goto main_exit;
                    break;
                }
            }
        }
        printf("\rCommand probe loop completed.  %u acks found.\n", ecnt);
    }
    else if (scp_mode == 2)
    {
        // *** search for length ***
        printf("Length probe: from 0 to 16. Command: %02X.\n", scp_parm1);
        for (cnt = 0; cnt != 16; cnt++)
        {
            for (rpt = 3; rpt != 0; rpt--)
            {
                printf("\r Probe %02X... ", cnt);
                fflush(stdout);
                // break carrier
                if ((err = mfBrkCarrier(&ccr, 30)) != LERR_SUCCESS) goto main_exit;
                // set phd level
                if ((err = link_packet(&ccr, CMS_ISO, phdl | 0x20, NULL, 0, NULL, 0)) != LERR_SUCCESS) goto main_exit;
                // select card
                if ((err = mfSelect(&ccr, 7, PICC_REQALL, NULL, NULL, NULL, NULL)) != LERR_SUCCESS) goto main_exit;
                // set command
                txbuf[0] = (BYTE)scp_parm1; txbuf[1] = (BYTE)scp_parm2;
                rxbuf[0] = 0;
                if (((err = mfLong(&ccr, txbuf, 2, rxbuf, &bitlen)) != LERR_SUCCESS) && (err != LERR_FAIL)) goto main_exit;
                if ((err == LERR_SUCCESS) && (bitlen == 4) && (rxbuf[0] == 0x0A))
                {
                    printf("ACK correct!\n");
                }
                else continue;

                // * Success! Sending string now... *
                for (ecnt = 0; ecnt != 32; ecnt++) txbuf[ecnt] = (BYTE)ecnt;
                rxbuf[0] = 0;
                printf("Sending data [");
                fflush(stdout);
                for (ecnt = 0; ecnt <= cnt; ecnt++)
                {
                    if (ecnt) printf(".%02X", txbuf[ecnt]);
                    else printf("%02X", txbuf[ecnt]);
                }
                printf("]... ");
                fflush(stdout);
                if (((err = mfLong(&ccr, txbuf, (BYTE)cnt + 1, rxbuf, &bitlen)) != LERR_SUCCESS) && (err != LERR_FAIL)) goto main_exit;
                if ((err == LERR_SUCCESS) && (bitlen >=4 ))
                {
                    bytelen = (bitlen % 8) ? bitlen / 8 + 1 : bitlen / 8;
                    strcpy(szLocal, "ACK=");
                    for (idx = 0; idx < bytelen; idx++)
                    {
                        if (idx) sprintf(&szLocal[strlen(szLocal)], ".");
                        sprintf(&szLocal[strlen(szLocal)], "%02X", rxbuf[idx]);
                    }
                    sprintf(&szLocal[strlen(szLocal)], "|%u bits", bitlen);
                    printf("%s\n", szLocal);
                    if ((err = mfBeep(&ccr, 0x02)) != LERR_SUCCESS) goto main_exit;
                    break;
                }
                else
                {
                    printf("\n");
                }

            }
        }
        printf("\nLength probe loop completed.\n");
    }

    err = LERR_SUCCESS;
main_exit:
    if (err)
    {
        verbose_err(err, szLocal);
        printf("%s\n", szLocal);
    }

#ifdef _RD0XAB_HIDRAW_H_
    if ((ccr.hL) && (err != LERR_HARDWARE) && (err != LERR_NOREADER))
#else
    if (((ccr.hL.u)||(ccr.hL.t)) && (err != LERR_HARDWARE) && (err != LERR_NOREADER))
#endif  // endif for subsytem info
    {
        if (err) mfBeep(&ccr, 0x03);    // if fail
        else mfBeep(&ccr, 0x02);        // if success
    }
    link_close(&ccr);
    return 0;
}
