// SRD mode demo (3-in-1), V1.6, hidraw/libusb-1.0 version, Linux

// _DEMO_MODE_ must be 0 or 1 or 2: 0 - SRD-ultra , 1 - SRD-auth, 2 - SRD-read
#define _DEMO_MODE_         0

#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include "rd0xAB.h"

#define PRG_VERSION         "1.6"

#ifdef _RD0XAB_HIDRAW_H_
#define PRG_COMMENT         " Hidraw version."
#else
#define PRG_COMMENT         " Libusb version."
#endif  // endif for subsytem info

#if _DEMO_MODE_ == 0        // 0 - read data from ultralight (without auth key)
//#pragma message("*** SRD-ULTRA demo compiling ***")
#define PRG_NAME            "srd-ultra"             // program name
#define START_MSG           "SRD-ultra for read data from ultralight card."
#define DEVM_CN             (1 + 8)                 // SRD(+1) mode + buzz_en(+8)
#define OPTS_CN             (4)                     // rwc_en(+4) + sto_en(+2) + polar(+1)
#define TMRS_CN             ((15 << 4) + 9)         // link_tout (0..15 << 4) + unlock_tout(0..15)
#define SRD_A_CN            0x00                    // AKSD: A - authorization command, 0x00 - none
#define SRD_K_CN            84                      // AKSD: K - AK number in reader memory, dummy value
#define SRD_S_CN            0                       // AKSD: S - start block/page for read
#define SRD_D_CN            0xF1                    // AKSD: D - Data descriptor: ((cnt-1)<<4) + A/D(+2)+(PG/BLK)(+1) - 16pgs

#elif _DEMO_MODE_ == 1      // 1 - authorization only, classic card; process uid, atqa, sak only (and auth key of course :) )
//#pragma message("*** SRD-AUTH demo compiling ***")
#define PRG_NAME            "srd-auth"              // program name
#define START_MSG           "SRD-auth demo, classic card authorization without read."
#define DEVM_CN             (1 + 8)                 // SRD(+1) mode + buzz_en(+8)
#define OPTS_CN             (4)                     // rwc_en(+4) + sto_en(+2) + polar(+1)
#define TMRS_CN             ((15 << 4) + 9)         // link_tout (0..15 << 4) + unlock_tout(0..15)
#define SRD_A_CN            PICC_AUTHENT1A          // AKSD: A - authorization command, PICC_AUTHENT1A = 0x60
#define SRD_K_CN            84                      // AKSD: K - AK number in reader memory
#define SRD_S_CN            0                       // AKSD: S - start block/page for read
#define SRD_D_CN            0x32                    // AKSD: D - Data descriptor: ((cnt-1)<<4) + A/D(+2)+(PG/BLK)(+1) - 4blks

#else                       // 2 - authorization and data read, classic card
//#pragma message("*** SRD-READ demo compiling ***")
#define PRG_NAME            "srd-read"              // program name
#define START_MSG           "SRD-read demo, read data from classic card."
#define DEVM_CN             (1 + 8)                 // SRD(+1) mode + buzz_en(+8)
#define OPTS_CN             (4)                     // rwc_en(+4) + sto_en(+2) + polar(+1)
#define TMRS_CN             ((15 << 4) + 9)         // link_tout (0..15 << 4) + unlock_tout(0..15)
#define SRD_A_CN            PICC_AUTHENT1A          // AKSD: A - authorization command, PICC_AUTHENT1A = 0x60
#define SRD_K_CN            84                      // AKSD: K - AK number in reader memory
#define SRD_S_CN            0                       // AKSD: S - start block/page for read
#define SRD_D_CN            0x30                    // AKSD: D - Data descriptor: ((cnt-1)<<4) + A/D(+2)+(PG/BLK)(+1) - 4blks
#endif  // endif for demo mode selector

// convert uidlen, atqa, sak to card type (copied from mif0xAB.c)
const char* acq2type_verb(DWORD acq)
{
    const char*      csType = "UNKN";

    if (acq == 0x07004400) csType = "UL";           // doub, 0044, 00 MIFARE Ultralight CL2
    else if (acq == 0x04000409) csType = "MINI";    // once, 0004, 09 MIFARE Mini

    else if (acq == 0x04000408) csType = "CL1K";    // once, 0004, 08 MIFARE 1K
    else if (acq == 0x04000488) csType = "CL1K";    // once, 0004, 18 MIFARE 1K
    else if (acq == 0x04000218) csType = "CL4K";    // once, 0002, 18 MIFARE 4K

    else if (acq == 0x07000404) csType = "CL1K1";   // once, 0004, 04 Any MIFARE CL1
    else if (acq == 0x07000204) csType = "CL4K1";   // once, 0002, 04 Any MIFARE CL1
    else if (acq == 0x07004404) csType = "CL1K1";   // once, 0044, 04 Any MIFARE CL1
    else if (acq == 0x07004204) csType = "CL4K1";   // once, 0042, 04 Any MIFARE CL1

    else if (acq == 0x07004408) csType = "PL2S1";   // doub, 0044, 08 MIFARE Plus CL2 2K Sec.lev.1
    else if (acq == 0x07004218) csType = "PL4S1";   // doub, 0042, 18 MIFARE Plus CL2 4K Sec.lev.1

    else if (acq == 0x04000410) csType = "PL2S2";   // once, 0004, 10 MIFARE Plus 2K Sec.lev.2
    else if (acq == 0x04000211) csType = "PL4S2";   // once, 0002, 11 MIFARE Plus 4K Sec.lev.2

    else if (acq == 0x07004410) csType = "PL2S2";   // doub, 0044, 10 MIFARE Plus CL2 2K Sec.lev.2
    else if (acq == 0x07004211) csType = "PL4S2";   // doub, 0042, 11 MIFARE Plus CL2 4K Sec.lev.2

    else if (acq == 0x04000420) csType = "PL2S3";   // once, 0004, 20 MIFARE Plus 2K Sec.lev.3
    else if (acq == 0x04000220) csType = "PL4S3";   // once, 0002, 20 MIFARE Plus 4K Sec.lev.3
    else if (acq == 0x07004420) csType = "PL2S3";   // doub, 0044, 20 MIFARE Plus 2K Sec.lev.3
    else if (acq == 0x07004220) csType = "PL4S3";   // doub, 0042, 20 MIFARE Plus 4K Sec.lev.3

    else if (acq == 0x07034420) csType = "DF2";     // doub, 0344, 20 MIFARE DESFire CL2
    else if (acq == 0x07034460) csType = "DF2";     // doub, 0344, 60 MIFARE DESFire card emulation
    else if ((acq&(~0xFF0000FF)) == 0x07000024) csType = "DF1";     // doub, xxxx, 24 MIFARE DESFire CL1

    else if ((acq&(~0x000F0000)) == 0x04000420) csType = "SM1K";    // once, 0x04, 20 SmartMX with MIFARE 1K emulation
    else if ((acq&(~0x000F0000)) == 0x04000428) csType = "SM1K";    // once, 0x04, 28 SmartMX | SAM)AV2 with MIFARE 1K emulation
    else if ((acq&(~0x000F0000)) == 0x04000468) csType = "SM1K";    // once, 0x04, 68 SmartMX with MIFARE 1K emulation

    else if ((acq&(~0x000F0000)) == 0x04000238) csType = "SM4K";    // once, 0x02, 38 SmartMX with MIFARE 4K emulation
    else if ((acq&(~0x000F0000)) == 0x07004820) csType = "SMMX";    // once, 0x48, 20 SmartMX

    return csType;
}

void verbose_err(DWORD err, char* szErr)
{
    if (err == LERR_SUCCESS) strcpy(szErr, "Success!");
    else if (err == LERR_FAIL) strcpy(szErr, "Command execution error!");
    else if (err == LERR_HARDWARE) strcpy(szErr, "Hardware error!");
    else if (err == LERR_TIMEOUT_TX) strcpy(szErr, "Communication error: TX timeout!");
    else if (err == LERR_TIMEOUT_RX) strcpy(szErr, "Communication error: RX timeout!");
    else if (err == LERR_NOTSYNC) strcpy(szErr, "Communication error: synchronization lost!");
    else if (err == LERR_NOREADER) strcpy(szErr, "Can't find reader!");
    else if (err == LERR_NOCARD) strcpy(szErr, "Can't find card!");
    else if (err == LERR_INVALIDCARD) strcpy(szErr, "Invalid card type!");
    else if (err == LERR_INVALIDKEY) strcpy(szErr, "Invalid auth key or card type!");
    else if (err == LERR_IOERR) strcpy(szErr, "I/O error!");
    else if (err == LERR_DATAREAD) strcpy(szErr, "Can't read data from card!");
    else if (err == LERR_DATAWRITE) strcpy(szErr, "Can't write data to card!");
    else if (err == LERR_BUSY) strcpy(szErr, "Device busy error!");
    else strcpy(szErr, "Unknown error!");
}

void usage(void)
{
    printf("Usage for USB: %s\n", PRG_NAME);
    printf("Usage for COM: %s /dev/tty2 9600\n", PRG_NAME);
}

BOOL _kbhit(void)
{
     struct   termios oldt, newt;
     int              ch;
     int              oldf;

     tcgetattr(STDIN_FILENO, &oldt);
     newt = oldt;
     newt.c_lflag &= ~(ICANON | ECHO);
     tcsetattr(STDIN_FILENO, TCSANOW, &newt);
     oldf = fcntl(STDIN_FILENO, F_GETFL, 0);
     fcntl(STDIN_FILENO, F_SETFL, oldf | O_NONBLOCK);

     ch = getchar();

     tcsetattr(STDIN_FILENO, TCSANOW, &oldt);
     fcntl(STDIN_FILENO, F_SETFL, oldf);

     if (ch != EOF)
     {
         ungetc(ch, stdin);
         return TRUE;
     }
     return FALSE;
}

// *** main ***
int  main(int argc, char *const argv[])
{
    CCR             ccr;
    char            *devtty = NULL;
    DWORD           cbr = 0;
    DWORD           err = LERR_SUCCESS;

    char            szLocal[32];
    BYTE            buf[78], key;
    DWORD           cnt, tmp;
    BOOL            brkf = FALSE;

    memset(&ccr, 0, sizeof(ccr));
    printf("%s COM/USB, V%s.%s\n--\n", START_MSG, PRG_VERSION, PRG_COMMENT);
    // Parse command line arguments
    if ((argc != 1) && (argc != 3))
    {
        usage();
        goto main_ferr;
    }
    if (argc == 3)
    {
        devtty = argv[1];
        sscanf(argv[2], "%u", &cbr);
    }
    if ((!devtty) && (argc > 1))
    {
        printf("Invalid parameters!\n\n");
        usage();
        goto main_ferr;
    }

    // if port != 0 then try open COM else try open USB
    if (devtty)
    {
        printf("COM: %s, CBR=%u. ", devtty, cbr);
        if ((err = link_comopen(devtty, cbr, &ccr))) goto main_ferr;
    }
    else
    {
        printf("USB. ");
        if ((err = link_hidopen(0, NULL, &ccr)))
        {
             err = LERR_NOREADER;
             goto main_ferr;
        }
    }
    // read device description: HW, NVMSZ, model, SN
    if ((err = link_packet(&ccr, CMG_VER, 0, NULL, 0, buf, 6)) != LERR_SUCCESS)
    {
         if ((devtty) && (err == LERR_HARDWARE)) err = LERR_NOREADER;
         goto main_ferr;
    }
    printf("Reader ");
    if ((buf[2] & 0xF0) == 0x10) printf("RD-01A");
    else if ((buf[2] & 0xF0) > 0x10) printf("RD-%02uAB", buf[2] / 16);
    else printf("???");
    printf(", SN:%08X, HW:%u.%u, %uKb.",
        (((DWORD)buf[2]) << 24) | (((DWORD)buf[3]) << 16) | (((DWORD)buf[4]) << 8) | ((DWORD)buf[5]),
        buf[0] >> 4, buf[0] & 15, ((DWORD)buf[1] + 1) / 2);
    if (buf[0] < 0x20)
    {
        printf(" ERROR: HW version must be >= 2.0!\n");
        goto main_ferr;
    }
    printf("\n--\n");

    // temporary set DIR mode, buzzer on
    if ((err = link_packet(&ccr, CMS_MOD, 3 + 8, NULL, 0, NULL, 0)) != LERR_SUCCESS) goto main_ferr;

    // *** print & parse setup parameters ***
    brkf = FALSE;   // TRUE if invalid settings
    // device mode
    printf("Mode ................... ");
    if ((DEVM_CN & 3) == 0) printf("ACS");
    else if ((DEVM_CN & 3) == 1) printf("SRD");
    else if ((DEVM_CN & 3) == 2) printf("IBE");
    else printf("DIR");
    if ((DEVM_CN & 3) != 1)
    {
        printf(" (invalid mode)");
        brkf = TRUE;
    }
    printf("\n");
    // buzzer
    printf("Buzzer ................. %s.\n", (DEVM_CN & 8) ? "enable" : "disable");
    // RW-flag
    printf("RW-card ................ %s.\n", (OPTS_CN & 4) ? "enable" : "disable");
    // polarity
    printf("Polarity ............... %s.\n", (OPTS_CN & 1) ? "inverted" : "direct");
    // response timeout
    printf("Response timeout ....... ");
    tmp = (DWORD)(TMRS_CN & 0xF0) >> 4;     // select link_timer nibble
    if (tmp == 0)
    {
        printf("without request (invalid).\n");
    }
    else if (tmp == 1) printf("request only.\n");
    else
    {
        // time = (ack_timer * 16 + 15) * 25mS
        printf("%.1fs.\n", (float)(tmp * 16 + 15) * 25 / 1000);
    }
    // unlock time
    printf("Unlock time ............ ");
    tmp = (DWORD)(TMRS_CN & 0x0F);          // select unlock_timer nibble
    if (tmp == 0)
    {
        printf("trigger mode.\n");
    }
    else
    {
        // time = (unlock_timer * 16 + 15) * 50mS
        printf("%.1fs.\n", (float)(tmp * 16 + 15) * 50 / 1000);
    }

    // authorization command
    printf("\nAuth command ........... %02X.\n", SRD_A_CN);
    // authorization key number (AK)
    printf("Auth key number ........ %u.\n", SRD_K_CN);
    // start block/page
    printf("Start block/page ....... %u.\n", SRD_S_CN);
    // auth/data read mode
    printf("Data/Auth mode ......... %s.\n", (SRD_D_CN & 2) ? "auth only" : "data read");
    // page mode
    printf("Block/page mode ........ %s.\n", (SRD_D_CN & 1) ? "page mode (4 bytes per unit)" : "block (16 bytes per unit)");

    // block's/page's counter : datalen = (sc_aksd[3] & 2) ? 0 : ((sc_aksd[3] & 1) ? 4 : 16) * ((sc_aksd[3] >> 4) + 1);
    if (SRD_D_CN & 2) tmp = 0;
    else if (SRD_D_CN & 1) tmp = 4 * ((SRD_D_CN >> 4) + 1);
    else tmp = 16 * ((SRD_D_CN >> 4) + 1);
    printf("Blocks/pages for read .. %u (%u bytes).\n", (SRD_D_CN >> 4) + 1, tmp);
    printf("--\n");
    // exit if invalid parameters
    if (brkf)
    {
        printf("Can't continue with current parameters! Change options then recompile and re-run program.\n");
        goto main_ferr;
    }

    printf("Setup parameters... ");
    // *** ROM setup; set if difference present only ***
    // buzzer & device mode
    if ((err = link_packet(&ccr, CMG_MOD, 0x80, NULL, 0, buf, 1)) != LERR_SUCCESS) goto main_ferr;
    if (buf[0] != DEVM_CN)
    {
        // set ROM mode
        if ((err = link_packet(&ccr, CMS_MOD, 0x80 | DEVM_CN, NULL, 0, NULL, 0)) != LERR_SUCCESS) goto main_ferr;
    }
    // options
    if ((err = link_packet(&ccr, CMG_OPT, 0x80, NULL, 0, buf, 1)) != LERR_SUCCESS) goto main_ferr;
    if (buf[0] != OPTS_CN)
    {
        // set new options
        if ((err = link_packet(&ccr, CMS_OPT, OPTS_CN, NULL, 0, NULL, 0)) != LERR_SUCCESS) goto main_ferr;
    }
    // timers
    if ((err = link_packet(&ccr, CMG_TMR, 0, NULL, 0, buf, 1)) != LERR_SUCCESS) goto main_ferr;
    if (buf[0] != TMRS_CN)
    {
        // set new timers value
        if ((err = link_packet(&ccr, CMS_TMR, TMRS_CN, NULL, 0, NULL, 0)) != LERR_SUCCESS) goto main_ferr;
    }
    // SRD parameters: AKSD
    if ((err = link_packet(&ccr, CMG_SRD, 0, NULL, 0, buf, 4)) != LERR_SUCCESS) goto main_ferr;
    if ((buf[0] != SRD_A_CN) || (buf[1] != SRD_K_CN) || (buf[2] != SRD_S_CN) || (buf[3] != SRD_D_CN))
    {
        buf[0] = SRD_A_CN; buf[1] = SRD_K_CN; buf[2] = SRD_S_CN; buf[3] = SRD_D_CN;
        if ((err = link_packet(&ccr, CMS_SRD, 0, buf, 4, NULL, 0)) != LERR_SUCCESS) goto main_ferr;
    }

    // final beep :-)
    if ((err = link_packet(&ccr, CMF_BUZ, 0x02, NULL, 0, NULL, 0)) != LERR_SUCCESS) goto main_ferr;
    // set current device mode (ROM -> RAM)
    if ((err = link_packet(&ccr, CMS_MOD, DEVM_CN, NULL, 0, NULL, 0)) != LERR_SUCCESS) goto main_ferr;
    printf("Ok.\n");

    // *** main loop ***
    printf("--\nPress 1 for lock, 2 for unlock, 3 for default action.\nPress Esc for exit.\n\nWait for card...\n");
    while (!brkf)
    {
        DWORD           stoptick;
        time_t          systm;
        struct   tm     *ptm;

        if (_kbhit())
        {
            if ((key = (BYTE)getchar()) == 0) key = (BYTE)getchar();
            if (key == 0x1B)
            {
                printf("^Esc pressed!\n");
                brkf = TRUE;
                break;
            }
        }

        // calculate length of data: datalen = (sc_aksd[3] & 2) ? 0 : ((sc_aksd[3] & 1) ? 4 : 16) * ((sc_aksd[3] >> 4) + 1);
        if (SRD_D_CN & 2) tmp = 0;
        else if (SRD_D_CN & 1) tmp = 4 * ((SRD_D_CN >> 4) + 1);
        else tmp = 16 * ((SRD_D_CN >> 4) + 1);

        // wait for unlock request from reader: length = 14b{1b(uidlen) + 10b(uid) + 1b(sak) + 2b(atqa)} + datalen{blks|pgs}
        if ((err = link_rqwait(&ccr, buf, (BYTE)(SZ_SCCD + tmp))) == LERR_HARDWARE) goto main_ferr;
        if ((err != LERR_SUCCESS) || ((buf[0] & 0x0F) > SZ_UID)) continue;  // uidlen <= 10, b.7 - RW-flag

        // build type string
        strcpy(szLocal, acq2type_verb( ((DWORD)(buf[0] & 0x0F) << 24) |                 // uidlen
            (((DWORD)buf[12]) << 16) | (((DWORD)buf[13]) << 8) | ((DWORD)buf[11])));    // ATQA + SAK
        if (buf[0] & MFDM_RWC) strcat(szLocal, "-RW");

        // *** print request message ***
        time(&systm); ptm = localtime(&systm);
        printf("[%2u/%02u/%02u %2u:%02u:%02u]",
            ptm->tm_mday, ptm->tm_mon + 1, ptm->tm_year % 100,
            ptm->tm_hour, ptm->tm_min, ptm->tm_sec); // print time
        // print card info: type, uid, data
        if (tmp)    // if datalen != 0
        {
            printf("\n    %s", szLocal);   // print card type
            // build UID string
            for (szLocal[0] = 0, cnt = 0; cnt < (DWORD)(buf[0] & 0x0F); cnt++)
                sprintf(&szLocal[strlen(szLocal)], "%02X", buf[cnt + 1]);
            printf(", UID:%s.", szLocal);   // print UID
            // print data from card
            for (cnt = 0; cnt != tmp; cnt++)
            {
                if ((cnt % 16) == 0) printf("\n    ");
                printf("%02X", buf[SZ_SCCD + cnt]);
            }
        }
        else        // if card parameters only
        {
            printf(" %8s", szLocal);    // print time, card type
            // build UID string
            for (szLocal[0] = 0, cnt = 0; cnt < (DWORD)(buf[0] & 0x0F); cnt++)
                sprintf(&szLocal[strlen(szLocal)], "%02X", buf[cnt + 1]);
            printf(", UID:%14s.", szLocal);     // print UID
        }
        // print final string
        printf(" Unlock? (1|2|3): ");
        fflush(stdout);
        // loop for getting acknowledge from user
        key = 0;
        stoptick = GetTickCount() + ((TMRS_CN & 0xF0) + 15) * 25 - 100; // count timeout time (100mS for IO by USB/COM)
        while (GetTickCount() < stoptick)
        {
            if (_kbhit())
            {
                if ((key = (BYTE)getchar()) == 0) key = (BYTE)getchar();
                if ((key == '1') || (key == '2') || (key == '3') || (key == 0x1B)) break;
            }
        }
        if (key == 0x1B) brkf = TRUE;
        if ((key != '1') && (key != '2') && (key != '3')) printf("none.\n");
        else
        {
            if (key == '1') printf("lock");
            else if (key == '2') printf("unlk");
            else printf("dflt");
            printf("...");
            key = (key & 3) - 1;
            // *** sending ack ***
            // 0 - no, 1 - yes, 2 - by reader db
            if ((err = link_rqack(&ccr, key)) == LERR_HARDWARE) goto main_ferr;
            if (err == LERR_SUCCESS) printf(" Ok.\n");
            else printf(" Fail.\n");
        }
    }
main_ferr:
    link_close(&ccr);
    if ((err) && (!brkf))
    {
        verbose_err(err, szLocal);
        printf("%s\n", szLocal);
    }
    return 0;
}
